<?php

namespace Drupal\config_sync;

use Drupal\config_snapshot\ConfigSnapshotStorageTrait;
use Drupal\config_sync\Plugin\SyncConfigCollectorInterface;
use Drupal\config_update\ConfigListInterface as ConfigUpdateListerInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Extension\Extension;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Provides methods related to listing configuration changes.
 */
class ConfigSyncLister implements ConfigSyncListerInterface {

  use ConfigSnapshotStorageTrait;
  use ConfigSyncExtensionsTrait;
  use StringTranslationTrait;

  /**
   * The configuration collector.
   *
   * @var \Drupal\config_sync\Plugin\SyncConfigCollectorInterface
   */
  protected $configCollector;

  /**
   * The configuration update lister.
   *
   * @var \Drupal\config_update\ConfigListInterface
   */
  protected $configUpdateLister;

  /**
   * The active configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * The provider configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $providerStorage;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * List of current config entity type labels, keyed by entity type.
   *
   * This is not set up until ConfigLister::listTypes() has been called.
   *
   * @var array
   */
  protected $configTypes = [];

  /**
   * Constructs a ConfigSyncLister object.
   *
   * @param \Drupal\config_sync\Plugin\SyncConfigCollectorInterface $config_collector
   *   The config collector.
   * @param \Drupal\config_update\ConfigListInterface $config_update_lister
   *   The configuration update lister.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active storage.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The provider configuration storage.
   * @param \Drupal\Core\Config\ConfigManagerInterface $config_manager
   *   The configuration manager.
   */
  public function __construct(SyncConfigCollectorInterface $config_collector, ConfigUpdateListerInterface $config_update_lister, StorageInterface $active_storage, StorageInterface $provider_storage, ConfigManagerInterface $config_manager) {
    $this->configCollector = $config_collector;
    $this->configUpdateLister = $config_update_lister;
    $this->activeStorage = $active_storage;
    $this->providerStorage = $provider_storage;
    $this->configManager = $config_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtensionChangelists(array $extension_names = []) {
    $changelist = [];
    // If no extensions were specified, use all installed extensions.
    if (empty($extension_names)) {
      $extension_names = $this->getSyncExtensions();
    }
    foreach ($extension_names as $type => $names) {
      foreach ($names as $name) {
        if ($extension_changelist = $this->getExtensionChangelist($type, $name)) {
          $changelist[$type][$name] = $extension_changelist;
        }
      }
    }

    return $changelist;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtensionChangelist($type, $name) {
    $snapshot_storage = $this->getConfigSnapshotStorage(ConfigSyncSnapshotterInterface::CONFIG_SNAPSHOT_SET, $type, $name);

    // Set both storages to normalize on read to minimize spurious differences.
    if (($snapshot_normalize = $snapshot_storage->getNormalizeOnRead()) === FALSE) {
      $snapshot_storage->setNormalizeOnRead(TRUE);
    }
    if (($provider_normalize = $this->providerStorage->getNormalizeOnRead()) === FALSE) {
      $this->providerStorage->setNormalizeOnRead(TRUE);
    }

    // Create a storage with configuration installable from this extension.
    $pathname = $this->drupalGetFilename($type, $name);
    $extension = new Extension(\Drupal::root(), $type, $pathname);
    $extensions = [
      $name => $extension,
    ];
    $this->configCollector->addInstallableConfig($extensions);

    // Set up a storage comparer.
    $storage_comparer = new StorageComparer(
      $this->providerStorage,
      $snapshot_storage,
      $this->configManager
    );

    $return = [];

    if ($storage_comparer->createChangelist()->hasChanges()) {
      foreach ($storage_comparer->getAllCollectionNames() as $collection) {
        $changelist = $storage_comparer->getChangelist(NULL, $collection);
        // We're only concerned with create and update lists.
        unset($changelist['delete']);
        unset($changelist['rename']);
        $changelist = array_filter($changelist);

        // Convert the changelist into a format that includes the item label.
        foreach ($changelist as $change_type => $item_names) {
          foreach ($item_names as $item_name) {
            $adjusted_change_type = $change_type;
            // Check to ensure this item exists in the active configuration. If
            // not, convert this to a 'create' change. This may occur, for
            // example, where optional configuration that previously did not
            // meet the criteria for installation now does.
            if ($change_type === 'update' && !$this->activeStorage->exists($item_name)) {
              $adjusted_change_type = 'create';
            }
            // Figure out what type of config it is, and get the ID.
            $config_type = $this->configUpdateLister->getTypeNameByConfigName($item_name);

            if (!$config_type) {
              // This is simple config.
              $label = $item_name;
            }
            else {
              $config = $this->providerStorage->read($item_name);
              $definition = $this->configUpdateLister->getType($config_type);
              $key = $definition->getKey('label') ?: $definition->getKey('id');
              $label = $config[$key];
            }
            $return[$collection][$adjusted_change_type][$item_name] = $label;
          }
        }
      }
    }

    // Restore the prior normalize on read setting.
    $snapshot_storage->setNormalizeOnRead($snapshot_normalize);
    $this->providerStorage->setNormalizeOnRead($provider_normalize);

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function listConfigTypes() {
    if (empty($this->configTypes)) {
      $definitions = $this->configUpdateLister->listTypes();
      $config_types = array_map(function (EntityTypeInterface $definition) {
        return $definition->getLabel();
      }, $definitions);
      $config_types['system_simple'] = $this->t('Simple configuration');
      // Sort the entity types by label.
      uasort($config_types, 'strnatcasecmp');
      $this->configTypes = $config_types;
    }
    return $this->configTypes;
  }

  /**
   * Wraps the function drupal_get_filename().
   *
   * @param $type
   *   The type of the item; one of 'core', 'profile', 'module', 'theme', or
   *   'theme_engine'.
   * @param $name
   *   The name of the item for which the filename is requested. Ignored for
   *   $type 'core'.
   * @param $filename
   *   (Optional) The filename of the item if it is to be set explicitly rather
   *   than by consulting the database.
   *
   * @return
   *   The filename of the requested item or NULL if the item is not found.
   */
  protected function drupalGetFilename($type, $name, $filename = NULL) {
    return drupal_get_filename($type, $name, $filename);
  }

}
